@@ -25,9 +25,13 @@
 #include "ttywatch.h"
 
 void
-telnet_negotiate(int socket);
+telnet_negotiate_server(int socket);
+void
+telnet_negotiate_client(int socket);
 int
 telnet_process_input(net_conn *nc, char *data, int len);
+int
+telnet_data_has_break(net_conn *nc, char *data, int len);
 void
 telnet_send_output(net_conn *nc, char *data, int len);
 
