@@ -69,19 +69,41 @@ static struct poptOption optionsTable[] = {
     { "module", 'm', POPT_ARG_STRING, &modname, 'm',
       "path to module for logging this port", "path" },
     { "daemon", 'd', POPT_ARG_NONE, &do_daemon, 0,
-      "directory path for logfiles", "path" },
+      "run in background as a daemon", NULL },
       POPT_AUTOHELP
     { NULL, 0, 0, NULL, 0 }
 };
 
 static GSList *machines;
+static GSList *retry;    /* list of unconnected sockets to retry later */
 static GHashTable *machine_names;
 static GHashTable *machine_fds;
 static GHashTable *accept_fds;
 static GHashTable *net_fds;
+static GHashTable *waiting_fds;
+
+static int
+machine_name_is_net(char *name) {
+	char *colon, *c;
+
+	if (strchr(name, '/'))
+	    return 0;
+	colon = strchr(name, ':');
+	if (!colon)
+	    return 0;
+
+	/* are there only numerics following the colon? */
+	for (c = ++colon; *c; c++) {
+	    if (*c < '0' || *c > '9')
+		return 0;
+	}
+
+	return 1;
+}
 
 /* get machine* for machine name, creating a new machine* if necessary */
-machine *get_machine(char *mach) {
+static machine *
+get_machine(char *mach) {
 	machine *m;
 
 	m = g_hash_table_lookup(machine_names, mach);
@@ -90,14 +112,23 @@ machine *get_machine(char *mach) {
 	m = calloc(sizeof(machine), 1);
 	if (!m) abort();
 
+	m->dev_nc.fd = -1;
+	m->dev_nc.m = m;
+	m->dev_nc.ts = TS_DATA;
+
+	m->accept_fd = -1;
 	m->name = strdup(mach);
 	m->speed = B115200;
 	m->logspew = 80*50; /* up to at least two screen pages */
 
-	/* defaults to devpath = /dev/<name> for symlinks */
-	m->devpath = malloc(strlen(mach)+6);
-	strcpy(m->devpath, "/dev/");
-	strcat(m->devpath, mach);
+	/* defaults to devpath = /dev/<name> for symlinks unless netname */
+	if (machine_name_is_net(mach)) {
+	    m->devpath = strdup(mach);
+	} else {
+	    m->devpath = malloc(strlen(mach)+6);
+	    strcpy(m->devpath, "/dev/");
+	    strcat(m->devpath, mach);
+	}
 
 	g_hash_table_insert(machine_names, m->name, m);
 	machines = g_slist_prepend(machines, m);
@@ -105,7 +136,8 @@ machine *get_machine(char *mach) {
 	return m;
 }
 
-speed_t symbolic_speed(int speednum) {
+static speed_t
+symbolic_speed(int speednum) {
    if (speednum >= 460800) return B460800;
    if (speednum >= 230400) return B230400;
    if (speednum >= 115200) return B115200;
@@ -127,7 +159,8 @@ speed_t symbolic_speed(int speednum) {
    return B50;
 }
 
-void setup_pidfile(void) {
+static void
+setup_pidfile(void) {
 	if (pidfile) {
 		char pid[32]; /* more than enough for an int... */
 		int pidfd;
@@ -139,77 +172,107 @@ void setup_pidfile(void) {
 	}
 }
 
-void delete_pidfile(void) {
+static void
+delete_pidfile(void) {
 	if (pidfile) {
 		unlink(pidfile);
 	}
 }
 
+static void
+add_to_retry_list(machine *m) {
+    m->ss = SS_TIMEOUT;
+    m->timeout_expires = time(NULL) + 1;
+    retry = g_slist_append(retry, m);
+}
+
+/* IFF the fd is a real device, send a BREAK down the line */
+static void
+send_break (net_conn *nc) {
+    if (nc->m->ss == SS_NOTASOCKET)
+	tcsendbreak(nc->fd, 0);
+}
+
 /* open and set termios settings for a single port, storing original settings
  * also set up ip port listening if requested
  */
-void setup_one_device(gpointer data, gpointer ignore) {
+static void
+setup_one_device(gpointer data, gpointer ignore) {
 	machine *m = data;
 
 	open_one_logfile(m);
 
-	m->dev_fd = open(m->devpath, O_RDWR|O_NOCTTY);
-	if (m->dev_fd < 0) {
-		warn(m, strerror(errno));
-		g_hash_table_remove(machine_names, m->name);
-		machines = g_slist_remove(machines, m);
-		free(m->devpath);
-		free(m->name);
-		free(m);
-		return;
+	if (machine_name_is_net(m->devpath)) {
+	    /* connection may not be available, but we keep trying */
+	    m->dev_nc.fd = socket_connect(m);
+	    if (m->dev_nc.fd > -1 && m->ss == SS_INPROGRESS) {
+		g_hash_table_insert(waiting_fds, &m->dev_nc.fd, m);
+	    }
+	} else {
+	    m->dev_nc.fd = open(m->devpath, O_RDWR|O_NOCTTY);
 	}
-	g_hash_table_insert(machine_fds, &m->dev_fd, m);
+	if (m->dev_nc.fd < 0) {
+	    warn(m, strerror(errno));
+	    g_hash_table_remove(machine_names, m->name);
+	    machines = g_slist_remove(machines, m);
+	    free(m->devpath);
+	    free(m->name);
+	    free(m);
+	    return;
+	}
+	g_hash_table_insert(machine_fds, &m->dev_nc.fd, m);
 
-	/* modify the port configuration */
-	tcgetattr(m->dev_fd, &m->ots);
-	m->ts = m->ots;
-	/* lots of random junk... */
-	m->ts.c_lflag &= ~ICANON;
-	m->ts.c_lflag &= ~(ECHO | ECHOCTL | ECHONL);
-	m->ts.c_cflag |= HUPCL;
-	m->ts.c_cflag &= ~CRTSCTS;
-	m->ts.c_iflag &= ~(IXON | IXOFF | IXANY);
-	m->ts.c_cc[VMIN] = 1;
-	m->ts.c_cc[VTIME] = 0;
+	if (m->ss == SS_NOTASOCKET) {
+	    /* modify the port configuration */
+	    tcgetattr(m->dev_nc.fd, &m->ots);
+	    m->ts = m->ots;
+	    /* lots of random junk... */
+	    m->ts.c_lflag &= ~ICANON;
+	    m->ts.c_lflag &= ~(ECHO | ECHOCTL | ECHONL);
+	    m->ts.c_cflag |= HUPCL;
+	    m->ts.c_cflag &= ~CRTSCTS;
+	    m->ts.c_iflag &= ~(IXON | IXOFF | IXANY);
+	    m->ts.c_cc[VMIN] = 1;
+	    m->ts.c_cc[VTIME] = 0;
 
-	/* throw away ^M (\r) characters on input */
-	m->ts.c_iflag |= IGNCR;
+	    /* throw away ^M (\r) characters on input */
+	    m->ts.c_iflag |= IGNCR;
 
-	/* set ALL the speeds */
-	cfsetospeed(&m->ts, m->speed);
-	cfsetispeed(&m->ts, m->speed);
+	    /* set ALL the speeds */
+	    cfsetospeed(&m->ts, m->speed);
+	    cfsetispeed(&m->ts, m->speed);
 
-	/* Now set the modified termios settings */
-	tcsetattr(m->dev_fd, TCSANOW, &m->ts);
+	    /* Now set the modified termios settings */
+	    tcsetattr(m->dev_nc.fd, TCSANOW, &m->ts);
+	}
 
 	/* Are we supposed to export this port via the network? */
 	if (m->ipport) {
 	    m->accept_fd = socket_listen(m->ipport, m->ipaddr);
-	    if (m->accept_fd) {
+	    if (m->accept_fd > -1) {
 		g_hash_table_insert(accept_fds, &m->accept_fd, m);
 	    }
 	}
 }
 
 /* set termios settings for all ports */
-void setup_devices(void) {
+static void
+setup_devices(void) {
 	g_slist_foreach(machines, setup_one_device, NULL);
 }
 
 /* restore termios settings for a single port, then close all fds */
-void cleanup_one_device(gpointer data, gpointer ignore) {
+static void
+cleanup_one_device(gpointer data, gpointer ignore) {
 	machine *m = data;
 
-	tcsetattr(m->dev_fd, TCSANOW, &m->ots);
-	close(m->dev_fd);
+	if (m->ss == SS_NOTASOCKET) {
+	    tcsetattr(m->dev_nc.fd, TCSANOW, &m->ots);
+	}
+	close(m->dev_nc.fd);
 	/* no logfile if a module is being used */
-	if (m->log_fd) close(m->log_fd);
-	if (m->accept_fd) close(m->accept_fd);
+	if (m->log_fd > -1) close(m->log_fd);
+	if (m->accept_fd > -1) close(m->accept_fd);
 	if (m->net_conns) {
 	    GSList *l;
 	    for (l = m->net_conns; l; l = g_slist_next(l)) {
@@ -223,19 +286,21 @@ void cleanup_one_device(gpointer data, gpointer ignore) {
 	}
 }
 
-void
+static void
 reopen_logfiles(int signo) {
     g_slist_foreach(machines, reopen_one_logfile, NULL);
 }
 
 /* restore all original terminal settings on exit */
-void cleanup_devices_and_exit(int signal) {
+static void
+cleanup_devices_and_exit(int signal) {
 	g_slist_foreach(machines, cleanup_one_device, NULL);
 	delete_pidfile();
 	exit(0);
 }
 
-struct pollfd * prepare_poll(int *nfds /*OUT*/, struct pollfd *p) {
+static struct pollfd *
+prepare_poll(int *nfds /*OUT*/, struct pollfd *p) {
 	GSList *l, *n;
 	machine *m;
 	int i;
@@ -244,8 +309,9 @@ struct pollfd * prepare_poll(int *nfds /*OUT*/, struct pollfd *p) {
 
 	for (l = machines; l; l = g_slist_next(l)) {
 	    m = l->data;
-	    (*nfds)++; /* one for the port */
-	    if (m->accept_fd)
+	    if (m->dev_nc.fd > -1)
+		(*nfds)++; /* one for the port */
+	    if (m->accept_fd > -1)
 		(*nfds)++; /* one for the accept fd */
 	    if (m->net_conns)
 		for (n = m->net_conns; n; n = g_slist_next(n))
@@ -256,10 +322,18 @@ struct pollfd * prepare_poll(int *nfds /*OUT*/, struct pollfd *p) {
 
 	for (l = machines, i=0; l; l = g_slist_next(l)) {
 	    m = l->data;
-	    p[i].fd = m->dev_fd;
-	    D("adding dev_fd %d\n", p[i].fd);
-	    p[i++].events = POLLIN;
-	    if (m->accept_fd) {
+	    if (m->dev_nc.fd > -1) {
+		p[i].fd = m->dev_nc.fd;
+		D("adding dev_nc.fd %d\n", p[i].fd);
+		if (m->ss == SS_INPROGRESS) {
+		    /* wait for the connection */
+		    p[i++].events = POLLOUT;
+		} else  {
+		    /* wait for data */
+		    p[i++].events = POLLIN;
+		}
+	    }
+	    if (m->accept_fd > -1) {
 		p[i].fd = m->accept_fd;
 		D("adding accept_fd %d\n", p[i].fd);
 		p[i++].events = POLLIN;
@@ -279,7 +353,7 @@ struct pollfd * prepare_poll(int *nfds /*OUT*/, struct pollfd *p) {
 }
 
 /* A network client has gone away */
-void
+static void
 remove_client (net_conn *nc, struct pollfd **ufds, int *nfds) {
 	machine *m;
 
@@ -292,7 +366,8 @@ remove_client (net_conn *nc, struct pollfd **ufds, int *nfds) {
 }
 
 
-machine *parse_argument (char c, machine *m) {
+static machine *
+parse_argument (char c, machine *m) {
 	int	conf_fd = 0;
 	struct stat sb;
 
@@ -354,7 +429,8 @@ machine *parse_argument (char c, machine *m) {
 				confnext = strchr(confline, '\n');
 				if (confnext) *confnext++ = '\0';
 
-				if (confline[0] == '#') continue;
+				if (!confline[0] || confline[0] == '#')
+				    continue;
 
 				poptParseArgvString(confline, &confc, (const char ***)&confv);
 				optCon = poptGetContext("ttywatch", confc, (const char **)confv,
@@ -370,7 +446,8 @@ machine *parse_argument (char c, machine *m) {
 	return m;
 }
 
-int main(int argc, char **argv) {
+int
+main(int argc, char **argv) {
 	char    c;            /* used for argument parsing */
 	struct sigaction sact;/* used to initialize the signal handler */
 	poptContext optCon;   /* context for parsing command-line options */
@@ -384,6 +461,7 @@ int main(int argc, char **argv) {
 	machine_fds = g_hash_table_new(g_int_hash, g_int_equal);
 	accept_fds = g_hash_table_new(g_int_hash, g_int_equal);
 	net_fds = g_hash_table_new(g_int_hash, g_int_equal);
+	waiting_fds = g_hash_table_new(g_int_hash, g_int_equal);
 
 	if (getuid())
 		logpath = ".";
@@ -428,13 +506,34 @@ int main(int argc, char **argv) {
 	do {
 	    int i, n, r;
 
-	    r = poll(ufds, nfds, -1);
+	    r = poll(ufds, nfds, retry ? 1000 : -1);
 	    if ((r < 0) && (errno != EINTR)) die("poll failed unexpectedly\n");
 	    for (n = 0; n < nfds; n++) {
 #		define TTYWATCH_BUFSIZE 1024
 		char buf[TTYWATCH_BUFSIZE];
 		machine *m;
 		net_conn *nc;
+		time_t now;
+
+		if (retry) {
+		    now = time(NULL);
+		    while (retry) {
+			m = retry->data;
+			if (m->timeout_expires < now) {
+			    m->dev_nc.fd = -1;
+			    socket_connect(m);
+			    if (m->dev_nc.fd > -1) {
+				if (m->ss == SS_INPROGRESS) {
+				    g_hash_table_insert(waiting_fds, &m->dev_nc.fd, m);
+				}
+				retry = g_slist_remove(retry, m);
+			    } else {
+				/* try again significantly later */
+				m->timeout_expires = time(NULL) + 10;
+			    }
+			}
+		    }
+		}
 
 		if (!ufds[n].revents) continue;
 
@@ -445,9 +544,13 @@ int main(int argc, char **argv) {
 			remove_client (nc, &ufds, &nfds);
 			break;
 		    } else {
-			/* If tcp monitored ports are later added, reopen
-			 * closed ports here
-			 */
+			if ((m = g_hash_table_lookup(machine_fds, &ufds[n].fd))) {
+			    /* a network data source has closed/died */
+			    close(m->dev_nc.fd);
+			    m->dev_nc.fd = -1;
+			    add_to_retry_list(m);
+			    ufds = prepare_poll(&nfds, ufds); break;
+			}
 
 			/* should not happen, but if it does, we cannot recover;
 			 * we would go into an infinite poll loop instead.
@@ -457,12 +560,37 @@ int main(int argc, char **argv) {
 		    }
 		}
 
+		if (ufds[n].revents & POLLOUT) {
+		    /* must be a finished non-blocking connect */
+		    if (!(m = g_hash_table_lookup(waiting_fds, &ufds[n].fd))) {
+			die("Socket on fd %d for POLLOUT DNE\n", ufds[n].fd);
+		    }
+		    g_hash_table_remove(waiting_fds, &ufds[n].fd);
+
+		    socket_continue_connect(m);
+		    switch (m->ss) {
+		    case SS_CONNECTED:
+			break;
+		    case SS_TIMEOUT:
+			/* APPEND to list of timed out sockets to retry later */
+			m->timeout_expires = time(NULL) + 1;
+			retry = g_slist_append(retry, m);
+			break;
+		    default:
+			die("illegal state %d for %s\n", m->ss, m->name);
+		    }
+		    ufds = prepare_poll(&nfds, ufds); break;
+		}
+
 		if (ufds[n].revents & POLLIN) {
 		    /* is it incoming data from a port? */
 		    if ((m = g_hash_table_lookup(machine_fds, &ufds[n].fd))) {
-			i = read(m->dev_fd, buf, TTYWATCH_BUFSIZE-1);
+			i = read(m->dev_nc.fd, buf, TTYWATCH_BUFSIZE-1);
 			if (i >= 1) {
 			    buf[i+1] = '\0';
+			    if (m->ss != SS_NOTASOCKET) {
+				i = telnet_process_input(&m->dev_nc, buf, i);
+			    }
 			    logfile_write(m->log_fd, buf, i);
 			    if (m->process_line)
 				m->process_line(m, buf);
@@ -485,13 +613,14 @@ int main(int argc, char **argv) {
 			    if (!nc) {
 				close(conn);
 				warn(m, "out of memory for new connection");
+				break;
 			    }
 			    nc->m = m;
 			    nc->fd = conn;
 
 			    m->net_conns = g_slist_prepend(m->net_conns, nc);
 			    g_hash_table_insert(net_fds, &nc->fd, nc);
-			    telnet_negotiate(nc->fd);
+			    telnet_negotiate_server(nc->fd);
 			    /* if available, send some existing logfile data */
 			    logfile_spew(nc);
 			    ufds = prepare_poll(&nfds, ufds); break;
@@ -502,12 +631,20 @@ int main(int argc, char **argv) {
 			m = nc->m;
 			i = read(ufds[n].fd, buf, TTYWATCH_BUFSIZE-1);
 			if (!i) {
-			    /* why not POLLHUP? <sigh> */
+			    /* client connection closed (POLLHUP only for ttys, <sigh>) */
 			    remove_client (nc, &ufds, &nfds);
 			    break;
 			}
 			i = telnet_process_input(nc, buf, i);
-			write(m->dev_fd, buf, i);
+
+			if (telnet_data_has_break(nc, buf, i)) {
+			    send_break (nc);
+			}
+			if (m->ss == SS_NOTASOCKET) {
+			    write(m->dev_nc.fd, buf, i);
+			} else {
+			    telnet_send_output(&m->dev_nc, buf, i);
+			}
 			/* write(STDERR_FILENO, buf, i); */
 
 		    } else {
