@@ -1,7 +1,7 @@
 /* ttywatch.h -- log output from TTYs, based on robin.c from the book
  * Linux Application Development, by Michael K. Johnson and Erik W. Troan
  *
- * Copyright  2000 Michael K. Johnson <johnsonm@redhat.com>
+ * Copyright  2000, 2001 Michael K. Johnson <johnsonm@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@
 #define __TTYTWATCH_H__
 
 #include <termios.h>
+#include <time.h>
 #include <dlfcn.h>
 #include <glib.h>
 
@@ -43,6 +44,15 @@ struct net_conn {
 	telnet_state ts;
 };
 
+
+/* Outgoing socket connection state */
+typedef enum {
+	SS_NOTASOCKET = 0,
+	SS_INPROGRESS,
+	SS_CONNECTED,
+	SS_TIMEOUT,
+} socket_state;
+
 typedef struct machine machine;
 struct machine {
 	char *name;
@@ -52,8 +62,11 @@ struct machine {
 	int logspew;
 	void (*process_line)(machine *m, char *data);
 	void *modhandle;
-	int dev_fd;
-	int log_fd;
+	void *moddata;          /* reserved for module use */
+	net_conn dev_nc;        /* might be a plain file or a net connection */
+	socket_state ss;        /* describes dev_fd */
+	time_t timeout_expires; /* when to try again if ss == SS_TIMEOUT */
+	volatile int log_fd;    /* volatile because modified from signal handler */
 	int accept_fd;
 	GSList *net_conns;
 	speed_t speed;
